/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : Mon Aug 30 16:12:57 EDT 2004
    copyright            : (C) 2004 by Peter Donis
    email                : peterdonis@alum.mit.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <argp.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

/* Defines that we'll need for protocols */

#define FTP_ID "ftp"
#define FTP_PORT 21

#define HTTP_ID "http"
#define HTTP_PORT 80
#define HTTP_STRING "GET / HTTP/1.0\r\n\r\n\0"
#define HTTP_SIZE sizeof(HTTP_STRING) - 1

#define IMAP_ID "imap"
#define IMAP_PORT 143
#define IMAP_STRING "A1 CAPABILITY\r\n\0"
#define IMAP_SIZE sizeof(IMAP_STRING) - 1
#define LOGOUT_STRING "A2 LOGOUT\r\n\0"
#define LOGOUT_SIZE sizeof(LOGOUT_STRING) - 1

#define POP_ID "pop"
#define POP_PORT 110

#define NEWS_ID "news"
#define NEWS_PORT 119

#define SMTP_ID "smtp"
#define SMTP_PORT 25

#define CRLF_STRING "\r\n\0"
#define CRLF_SIZE sizeof(CRLF_STRING) - 1

#define QUIT_STRING "QUIT\r\n\0"
#define QUIT_SIZE sizeof(QUIT_STRING) - 1

/* Output messages */

#define WAITING " ...\n"
#define DNSONLY_MSG "Doing DNS lookup on %s...\n"
#define START_MSG "Contacting URL %s"
#define PROTO_MSG " using %s"
#define PORT_MSG " on port %i"
#define ADDRCORR_MSG "IP address for %s is %s.\n"
#define NAMECORR_MSG "Domain name for %s is %s.\n"
#define CONNECT_MSG "Connected ...\n"
#define SERVER_MSG "Server returned the following:\n\n"
#define CLOSE_MSG "Connection closed.\n"

/* Error messages */

#define ALLOCERR_MSG "Could not allocate memory.\n"
#define SOCKERR_MSG "Could not create socket.\n"
#define NAMEERR_MSG "Unknown domain name %s.\n"
#define ADDRERR_MSG "Unknown IP address %s.\n"
#define CONNECTERR_MSG "Could not connect to host %s.\n"
#define PROTERR_MSG "%s protocol %s not understood.\nKnown protocols are ftp, http, imap, news, pop, and smtp."
#define PORTERR_MSG "%s %s is not a valid port number."
#define URLERR_MSG "URL string is too long, cannot process."
#define OPTSNIPPET "in option switch"
#define URLSNIPPET "in URL"
#define ARGERR1_MSG "Too many arguments--you can only specify one URL.\n"
#define ARGERR2_MSG "Not enough arguments--no URL specified.\n"
#define DNSONLYERR_MSG "DNS lookup only; ignoring -p and -r switches.\n"

/* Protocol values for function that does the work */

#define PROTO_FTP FTP_PORT
#define PROTO_HTTP HTTP_PORT
#define PROTO_IMAP IMAP_PORT
#define PROTO_NEWS NEWS_PORT
#define PROTO_POP POP_PORT
#define PROTO_SMTP SMTP_PORT

/* Buffer lengths for parsing */

#define RETN_SIZE 1024
#define MAX_PROTO_ID_LEN 4
#define MAX_PORTNUM_LEN 5
#define MAX_URL_LEN 1024

/* URL delimiter lengths */

#define URL_PREFIX_LEN 3
#define URL_SUFFIX_LEN 1

/* Option switches */

#define SWITCH_LOOKUP 'l'
#define SWITCH_PROTO 'p'
#define SWITCH_PORT 'r'

/* String constants */

char *http_string = HTTP_STRING;
char *imap_string = IMAP_STRING;
char *logout = LOGOUT_STRING;
char *quitstr = QUIT_STRING;
char *linebreak = CRLF_STRING;
char *dbl_linebreak = "\r\n\r\n\0";
char *URLprefix = "://\0";
char *URLsuffix = ":\0";

/*** Internet connection code ***/

void output_buffer(char *buffer, char *endstr)
{
  char *bufend;
  
  /* Look for end of buffer and insert null after it to terminate string */
  bufend = strstr(buffer, endstr);
  if ( bufend != 0 ) {
    *(bufend + 2) = 0;
  }

  /* Output the string in the buffer */
  printf(buffer);
}

int get_info(int sock, int protocol)
{
  /* Allocate and clear buffer for server return data */
  char *get_buffer = (char *)malloc(RETN_SIZE+1);
  if ( get_buffer == 0 ) {
    printf(ALLOCERR_MSG);
    return EXIT_FAILURE;
  }
  memset(get_buffer, 0, RETN_SIZE+1);

  /* Start of output */
  printf(SERVER_MSG);

  /* Conversation with server */
  switch ( protocol ) {
    case PROTO_FTP:
    case PROTO_SMTP:
    case PROTO_POP:
    case PROTO_NEWS:
    case PROTO_IMAP: {
      /* Read welcome message from server and output */
      read(sock, get_buffer, RETN_SIZE);
      output_buffer(get_buffer, linebreak);

      if ( protocol == PROTO_IMAP ) {
        /* If IMAP send capability and logout messages, output replies */
        write(sock, imap_string, IMAP_SIZE);
        read(sock, get_buffer, RETN_SIZE);
        output_buffer(get_buffer, linebreak);
        write(sock, logout, LOGOUT_SIZE);
        read(sock, get_buffer, RETN_SIZE);
        output_buffer(get_buffer, linebreak);
  
      } else if ( protocol != PROTO_FTP ) {
        /* If news, SMTP, or POP send quit message, output reply */
        write(sock, quitstr, QUIT_SIZE);
        read(sock, get_buffer, RETN_SIZE);
        output_buffer(get_buffer, linebreak);
      }

      break;
    }
    case PROTO_HTTP: {
      /* Send HTTP request to server and get default page */
      write(sock, http_string, HTTP_SIZE);     
      read(sock, get_buffer, RETN_SIZE);

      /* Output the result */
      output_buffer(get_buffer, dbl_linebreak);

      break;
    }
  }

  /* End of output */
  printf("\n");

  /* This ensures that the output is displayed immediately (instead of waiting
     until the program has exited) when stdout is something other
     than a shell terminal (e.g., when we're running as a child of KIDserver)
     -- we'll do this several times at the major stages of the program */
  fflush(stdout);

  /* Free the allocated buffer! */
  free(get_buffer);

  return EXIT_SUCCESS;
}

int talk_to_URL(char *URL, int portnum, int protocol, int DNSonly)
{
  int retval;
  int sock;
  int actualprotocol;
  int actualport;
  struct sockaddr_in sock_name;
  struct hostent *hostinfo;

  /* Detect type of URL and process accordingly */
  if ( inet_aton(URL, &sock_name.sin_addr) ) {

    /* URL is an IP address. get domain name with reverse lookup and output */
    hostinfo = (struct hostent *)gethostbyaddr(&sock_name.sin_addr, sizeof(&sock_name.sin_addr), AF_INET);
    if ( hostinfo == NULL ) {
      printf(ADDRERR_MSG, URL);
      return EXIT_FAILURE;
    }
    printf(NAMECORR_MSG, URL, hostinfo->h_name);

  } else {

    /* Assume URL is a domain name, get IP address with DNS lookup and output */
    hostinfo = (struct hostent *)gethostbyname(URL);
    if ( hostinfo == NULL ) {
      printf(NAMEERR_MSG, URL);
      return EXIT_FAILURE;
    }
    sock_name.sin_addr = *(struct in_addr *) hostinfo->h_addr;
    printf(ADDRCORR_MSG, URL, inet_ntoa(sock_name.sin_addr));
  }

  /* We're done unless we're doing more than just a DNS lookup */
  if ( DNSonly ) {
      retval = EXIT_SUCCESS;
  } else {

    /* If no protocol specified, use HTTP */
    if ( protocol == 0 ) {
      actualprotocol = PROTO_HTTP;
    } else {
      actualprotocol = protocol;
    }

    /* If no port number specified, use default for protocol */
    if ( portnum == 0 ) {
      actualport = actualprotocol;
    } else {
      actualport = portnum;
    }

    /* Create Internet socket for TCP */
    sock = socket(PF_INET, SOCK_STREAM, 0);
    if ( sock < 0 ) {
      printf(SOCKERR_MSG);
      perror("socket");
      return EXIT_FAILURE;
    }

    sock_name.sin_family = AF_INET;
    sock_name.sin_port = htons(actualport);

    /* Connect to the URL */
    if ( connect(sock, (struct sockaddr *) &sock_name, sizeof(sock_name)) < 0 ) {
      printf(CONNECTERR_MSG, URL);
      perror("connect (client)");
      return EXIT_FAILURE;
    }
    printf(CONNECT_MSG);
    fflush(stdout);

    /* We're there, now get server data */
    retval = get_info(sock, actualprotocol);

    /* Close the socket */
    close(sock);
    printf(CLOSE_MSG);
  }

  /* We have output regardless of what we did or didn't do above, so flush it now */
  fflush(stdout);

  return retval;
}

/*** Command line parsing code ***/

int getprotocol(char *proto) {
  if ( strcmp(proto, FTP_ID) == 0 ) {
    return PROTO_FTP;
  } else if ( strcmp(proto, HTTP_ID) == 0 ) {
    return PROTO_HTTP;
  } else if ( strcmp(proto, IMAP_ID) == 0 ) {
    return PROTO_IMAP;
  } else if ( strcmp(proto, NEWS_ID) == 0 ) {
    return PROTO_NEWS;
  } else if ( strcmp(proto, POP_ID) == 0 ) {
    return PROTO_POP;
  } else if ( strcmp(proto, SMTP_ID) == 0 ) {
    return PROTO_SMTP;
  }
  return 0;
}

struct argp_switches
{
  char *URL;
  char *URLbody;
  char *proto;
  int portnum;
  int protocol;
  int DNSonly;
};

void parse_protocol(char *arg, struct argp_switches *switches, struct argp_state *state,
  char *err)
{
  switches->protocol = getprotocol(arg);
  if ( switches->protocol != 0 ) {
    switches->proto = arg;
  } else {
    /* Unknown protocol */
    argp_error(state, PROTERR_MSG, arg, err);
  }
}

void parse_portnum(char *arg, struct argp_switches *switches, struct argp_state *state,
  char *err)
{
  switches->portnum = atoi(arg);
  if ( switches->portnum == 0 ) {
    /* Invalid port number */
    argp_error(state, PORTERR_MSG, arg, err);
  }
}

/* Buffers for parsing URL */
char prefix[MAX_PROTO_ID_LEN+1];
char tmpbuf[MAX_URL_LEN+1];

void parse_url (struct argp_switches *switches, struct argp_state *state)
{
  /* Temporary variables for parsing URL */
  char *tmp;
  int tmpsize;

  /* If URL has prefix, separate and parse */
  switches->URLbody = switches->URL;
  tmp = strstr(switches->URLbody, URLprefix);
  if ( tmp != 0 ) {

    /* There was a prefix to the URL, get its protocol number */
    tmpsize = tmp - switches->URLbody;
    if ( tmpsize > MAX_PROTO_ID_LEN ) {
      /* Invalid protocol */
      argp_error(state, PROTERR_MSG, prefix, URLSNIPPET);
    } else {
      strncpy(prefix, switches->URLbody, tmpsize);
      prefix[tmpsize] = 0;
      parse_protocol(prefix, switches, state, URLSNIPPET);
    }

    /* Rest of the URL is after the prefix delimiter */
    switches->URLbody = tmp + URL_PREFIX_LEN;
  }

  /* If URL has suffix, separate and parse */
  tmp = strstr(switches->URLbody, URLsuffix);
  if ( tmp != 0 ) {

    /* There was a suffix to the URL, get the port number */
    parse_portnum(tmp + URL_SUFFIX_LEN, switches, state, URLSNIPPET);

    /* We'll have to copy the actual body of the URL into a buffer & point to it */
    tmpsize = strlen(switches->URLbody) - strlen(tmp);
    if ( tmpsize > MAX_URL_LEN ) {
      /* URL string is too long */
      argp_error(state, URLERR_MSG);
    } else {
      strncpy(tmpbuf, switches->URLbody, tmpsize);
      switches->URLbody = &tmpbuf[0];
    }
  }
}

error_t argp_parser(int key, char *arg, struct argp_state *state)
{
  struct argp_switches *switches = state->input;
  
  switch ( key ) {
    case SWITCH_LOOKUP:
      /* Only do DNS lookup */
      switches->DNSonly = 1;
      break;

    case SWITCH_PROTO:
      /* Argument should be a protocol */
      parse_protocol(arg, switches, state, OPTSNIPPET);
      break;

    case SWITCH_PORT:
      /* Argument should be a port number */
      parse_portnum(arg, switches, state, OPTSNIPPET);
      break;

    case ARGP_KEY_ARG:
      if ( state->arg_num > 0 ) {
        /* Too many arguments */
        printf(ARGERR1_MSG);
        argp_usage(state);
      } else {
        /* This is the URL argument */
        switches->URL = arg;
        parse_url(switches, state);
      }
      break;

    case ARGP_KEY_END:
      if ( state->arg_num < 1 ) {
        /* Not enough arguments */
        printf(ARGERR2_MSG);
        argp_usage(state);
      }
      break;

    default:
      return ARGP_ERR_UNKNOWN;
  }
  return 0;
}

/* argp data */

const char *argp_program_version = "idserver 0.6";
const char *argp_program_bug_address = "<peterdonis@alum.mit.edu>";
static struct argp_option options[] = {
  {"lookup", SWITCH_LOOKUP, 0, 0, "Only do DNS lookup, no server query"},
  {"protocol", SWITCH_PROTO, "PROTOCOL", 0, "Use the specified protocol to contact the server"},
  {"port", SWITCH_PORT, "PORT", 0, "Use the specified port number to contact the server"},
  { 0 }
};
static char args_doc[] = "URL";
static char doc[] = "IDserver: a command line utility for getting information from a web server.";
static struct argp argp_struct = {options, argp_parser, args_doc, doc};

int main(int argc, char *argv[])
{
  struct argp_switches switches;

  /* Default values */
  switches.URL = 0;
  switches.URLbody = 0;
  switches.proto = 0;
  switches.portnum = 0;
  switches.protocol = 0;
  switches.DNSonly = 0;

  /* Parse command line arguments */
  argp_parse(&argp_struct, argc, argv, 0, 0, &switches);

  /* Output opening message */
  if ( switches.DNSonly ) {
    printf(DNSONLY_MSG, switches.URLbody);
    if ( switches.proto || switches.portnum ) {
      /* Protocol and port number not used in this mode */
      printf(DNSONLYERR_MSG);
    }
  } else {
    printf(START_MSG, switches.URLbody);
    if ( switches.proto ) {
      printf(PROTO_MSG, switches.proto);
    }
    if ( switches.portnum ) {
      printf(PORT_MSG, switches.portnum);
    }
    printf(WAITING);
  }
  fflush(stdout);

  /* Do the work */  
  return talk_to_URL(switches.URLbody, switches.portnum, switches.protocol, switches.DNSonly);
}
